--
-- HardPointPlaceable
--
-- @author:    	Xentro (Marcus@Xentro.se)
-- @website:	www.Xentro.se
-- @history		v1.0   - 2015-03-27 - Initial implementations
-- 				v1.1   - 2015-09-12 - Fixed bug with trigger not getting emptied 
--				v1.1.1 - 2015-10-26 - Parts removed
-- 

source(g_currentModDirectory .. "Scripts/Events/SetAnimationEvent.lua");

HardPointPlaceable = {};
HardPointPlaceable.modPath = g_currentModDirectory;

HardPointPlaceable_mt = Class(HardPointPlaceable, Placeable);
InitObjectClass(HardPointPlaceable, "HardPointPlaceable");

function HardPointPlaceable:new(isServer, isClient, customMt)
	local mt = customMt;
	if mt == nil then
		mt = HardPointPlaceable_mt;
	end;
	
	local self = Placeable:new(isServer, isClient, mt);
	
    registerObjectClassName(self, "HardPointPlaceable");
	
	self.isEnabled = true;
	self.objectType = "placeableWorkshop";
	
	self.vehiclesInRange = {};
	self.numVehicles = 0;
	
	self.allowEnterWorkshop = true;
	self.buildingSize = 1;
	
	self.playerInRangeTool = {};
	self.clipAnimations = {};
	
	self.lightNodes = {};
	
    return self;
end;

function HardPointPlaceable:delete()
	if self.mapHotspot ~= nil then
		g_currentMission.ingameMap:deleteMapHotspot(self.mapHotspot);
	end;
    unregisterObjectClassName(self);
	
	for vehicle, v in pairs(self.vehiclesInRange) do
		self.vehiclesInRange[vehicle] = nil;
		vehicle:inHardPointTrigger(false);
	end;
	self.vehiclesInRange = nil;

	if g_currentMission.HardPoint ~= nil then
		g_currentMission.HardPoint:removeWorkshopTrigger(self);
	end;
	if g_currentMission.environment ~= nil then
		g_currentMission.environment:removeWeatherChangeListener(self);
	end;
	if g_currentMission.playerInRangeExtras ~= nil then
		g_currentMission.playerInRangeExtras[self] = nil;
	end;
	
	if self.trigger ~= nil then
		removeTrigger(self.trigger);
	end;
	
	HardPointPlaceable:superClass().delete(self);
end;

function HardPointPlaceable:deleteFinal()
	HardPointPlaceable:superClass().deleteFinal(self);
end;

function HardPointPlaceable:readStream(streamId, connection)
	HardPointPlaceable:superClass().readStream(self, streamId, connection);
	
	for k, anim in ipairs(self.clipAnimations) do
		self:setAnimationClip(k, streamReadBool(streamId), true);
	end;
end;

function HardPointPlaceable:writeStream(streamId, connection)
	HardPointPlaceable:superClass().writeStream(self, streamId, connection);
	
	for k, anim in ipairs(self.clipAnimations) do
		streamWriteBool(streamId, anim.isActivated);
	end;
end;

function HardPointPlaceable:load(xmlFilename, x,y,z, rx,ry,rz, moveMode, initRandom)
    if not HardPointPlaceable:superClass().load(self, xmlFilename, x, y, z, rx, ry, rz, moveMode, initRandom) then
		return false;
	end;
	
	local xmlFile = loadXMLFile("TempXML", xmlFilename);

	self.name = getXMLString(xmlFile, "placeable.HardPoint#name");
	if self.name ~= nil and g_i18n:hasText(self.name) then
		self.name = g_i18n:getText(self.name);
	end;
	
	self.buildingSize = Utils.getNoNil(getXMLInt(xmlFile, "placeable.HardPoint#buildingSize"), 3); -- small 1, medium 2, big 3
	self.playerDistance = Utils.getNoNil(getXMLFloat(xmlFile, "placeable.HardPoint#playerDistance"), 6);
	self.trigger = Utils.indexToObject(self.nodeId, getXMLString(xmlFile, "placeable.HardPoint#trigger"));
	self.lightNodes[1] = Utils.indexToObject(self.nodeId, getXMLString(xmlFile, "placeable.lights#corona"));
	self.lightNodes[2] = Utils.indexToObject(self.nodeId, getXMLString(xmlFile, "placeable.lights#real"));
	
	self.appearsOnPDA = Utils.getNoNil(getXMLBool(xmlFile, "placeable.HardPoint#appearsOnPDA"), true);
	if self.appearsOnPDA then
		local x, _, z = getWorldTranslation(self.trigger);
		self.mapHotspot = g_currentMission.ingameMap:createMapHotspot("Workshop", HardPointPlaceable.modPath .. "Shared/hud_pda_spot_workshop.png", x, z, nil, nil, false, false, true, 0, true);
	end;
	
	local i = 0;
	while true do
		local key = string.format("placeable.clipAnimations.animation(%d)", i);
		if not hasXMLProperty(xmlFile, key) then break; end;
		
		local node = Utils.indexToObject(self.nodeId, getXMLString(xmlFile, key .. "#node"));
		local locationNode = Utils.indexToObject(self.nodeId, getXMLString(xmlFile, key .. "#locationNode"));
		if node ~= nil and locationNode ~= nil then
			local charSet = getAnimCharacterSet(node);
			if charSet ~= nil then
				local animation = {};
				animation.node = node;
				animation.animCharSet = charSet;
				animation.clip = getAnimClipIndex(animation.animCharSet, getXMLString(xmlFile, key .. "#clipName"));
				animation.clipSpeed = Utils.getNoNil(getXMLFloat(xmlFile, key .. "#clipSpeed"), 1);
				animation.loop = Utils.getNoNil(getXMLBool(xmlFile, key .. "#loop"), false);
				animation.accelerationTime = Utils.getNoNil(getXMLFloat(xmlFile, key .. "#accelerationTime"), 0) / 1000;
				animation.decelerationTime = Utils.getNoNil(getXMLFloat(xmlFile, key .. "#decelerationTime"), 0) / 1000;
				animation.duration = getAnimClipDuration(animation.animCharSet, animation.clip);
				animation.canActivate = Utils.getNoNil(getXMLBool(xmlFile, key .. "#canActivate"), true);
				animation.isActivated = false;
				animation.isRunning = false;
				animation.currentAcc = 0;
				
				assignAnimTrackClip(animation.animCharSet, 0, animation.clip);
				setAnimTrackSpeedScale(animation.animCharSet, animation.clip, animation.clipSpeed);
				setAnimTrackLoopState(animation.animCharSet, 0, animation.loop);
				setAnimTrackTime(animation.animCharSet, 0, animation.currentTime);
				
				local input = getXMLString(xmlFile, key .. "#inputBinding");
				if input ~= nil then
					if InputBinding[input] ~= nil then
						animation.inputBinding = InputBinding[input];
					end;
				end;
				
				animation.text = {};
				local text = getXMLString(xmlFile, key .. "#l10nText");
				
				if text == nil then
					text = getXMLString(xmlFile, key .. "#l10nTextPositive");
					text2 = getXMLString(xmlFile, key .. "#l10nTextNegative");
					
					if text ~= nil and text2 ~= nil then
						if g_i18n:hasText(text) and g_i18n:hasText(text2) then 
							animation.text[2] = g_i18n:getText(text);
							animation.text[3] = g_i18n:getText(text2);
						end;
					end;
				else
					if g_i18n:hasText(text) then
						animation.text[1] = g_i18n:getText(text);
					end;
				end;
				
				local soundFile = getXMLString(xmlFile, key .. ".sound#file");
				if soundFile ~= nil and soundFile ~= "" then
					soundFile = Utils.getFilename(soundFile, self.baseDirectory);
					animation.soundVolume = Utils.getNoNil(getXMLFloat(xmlFile, key .. ".sound#volume"), 1.0);
					animation.soundRadius = Utils.getNoNil(getXMLFloat(xmlFile, key .. ".sound#radius"), 50);
					animation.soundInnerRadius = Utils.getNoNil(getXMLFloat(xmlFile, key .. ".sound#innerRadius"), 10);
					animation.soundFile = createAudioSource("soundFile" .. i, soundFile, animation.soundRadius, animation.soundInnerRadius, animation.soundVolume, 0);
					link(animation.node, animation.soundFile);
					setVisibility(animation.soundFile, false);
				end;
				
				self["doorDummyBool" .. i] = false;
				
				table.insert(self.playerInRangeTool, {node = locationNode, playerDistance = 4, interactiveData = {text1 = animation.text[3], text2 = animation.text[2], functionName = "inRangeCalledFunction", boolName = "doorDummyBool" .. i, offset = {0, 0.5, 0}}});
				table.insert(self.clipAnimations, animation);
				i = i + 1;
			else
				print("AnimationClips - Error: invalid animation node " .. node);
				break;
			end;
		else
			print("AnimationClips - Error: animation node is empty");
			break;
		end;
    end;
	
	delete(xmlFile);
	
    return true;
end;

function HardPointPlaceable:finalizePlacement()
	HardPointPlaceable:superClass().finalizePlacement(self);
	
	if self.trigger ~= nil then
		addTrigger(self.trigger, "triggerCallback", self);
	end;
	
	if g_currentMission.HardPoint ~= nil then
		local allow, txt = g_currentMission.HardPoint:checkBuildingSize(self.buildingSize);
		if not allow then
			g_currentMission.HardPoint:log(2, "Invalid buildingSize ( " .. self.buildingSize .. " ), it will be changed to 3 for " .. self.objectType .. " - " .. tostring(self.name));
			self.buildingSize = 3;
		end;
		
		g_currentMission.HardPoint:addWorkshopTrigger(self);
	end;
	
	if (self.lightNodes[1] ~= nil or self.lightNodes[2] ~= nil) then
		g_currentMission.environment:addWeatherChangeListener(self);
	end;
	
	if g_currentMission.playerInRangeExtras ~= nil then
		g_currentMission.playerInRangeExtras[self] = self;
	end;
end;

function HardPointPlaceable:loadFromAttributesAndNodes(xmlFile, key, resetVehicles)
	if not HardPointPlaceable:superClass().loadFromAttributesAndNodes(self, xmlFile, key, resetVehicles) then
		return false;
	end;
	
	if not resetVehicles then
		local i = 0;
		while true do
			local clip = key .. string.format(".clipAnimation(%d)", i);
			if not hasXMLProperty(xmlFile, clip) then break; end;
			
			local id = getXMLInt(xmlFile, clip .. "#id");
			if self.clipAnimations[id] ~= nil then
				local animTime = getXMLFloat(xmlFile, clip .. "#animationTime");
				local activated = getXMLBool(xmlFile, clip .. "#activated");
				
				if animTime > 0 then
					setAnimTrackTime(self.clipAnimations[id].animCharSet, self.clipAnimations[id].clip, animTime);
					self:setAnimationClip(id, activated, true);
				end;
			end;
			
			i = i + 1;
		end;
	end;
	
	return true;
end;

function HardPointPlaceable:getSaveAttributesAndNodes(nodeIdent)
	local attributes, nodes = HardPointPlaceable:superClass().getSaveAttributesAndNodes(self, nodeIdent);
	
	for k, anim in ipairs(self.clipAnimations) do
		if not anim.loop then
			if nodes ~= "" then
				nodes = nodes .. "\n";
			end;
			
			nodes = nodes .. nodeIdent .. '<clipAnimation';
			nodes = nodes .. ' id="' .. k .. '"';
			nodes = nodes .. ' animationTime="' .. string.format("%f", math.max(getAnimTrackTime(anim.animCharSet, anim.clip), 0)) .. '"';
			nodes = nodes .. ' activated="' .. tostring(anim.isActivated) .. '"';
			nodes = nodes .. " />"
		end;
	end;
	
	return attributes, nodes;
end;

function HardPointPlaceable:inRangeCalledFunction()
	if self.isClient then
		if g_currentMission.player ~= nil and g_currentMission.player.closestTool ~= nil then
			if g_currentMission.player.closestTool == self then
				local anim = self.clipAnimations[g_currentMission.player.closestToolLocationId];
				
				if anim ~= nil then
					self:setAnimationClip(g_currentMission.player.closestToolLocationId, not anim.isActivated);
				end;
			end;
		end;
	end;
end;

function HardPointPlaceable:update(dt)
	-- HardPointPlaceable:superClass().update(self, dt)
end;

function HardPointPlaceable:updateTick(dt)
	-- HardPointPlaceable:superClass().updateTick(self, dt)
	for k, v in ipairs(self.playerInRangeTool) do
		v.stopShowing = not self.clipAnimations[k].canActivate;
	end;
	
	if g_currentMission.HardPoint ~= nil then
		g_currentMission.HardPoint:updateVehicleCount(self);
	end;
	
	for k, anim in ipairs(self.clipAnimations) do
		if anim.isRunning then
			local currentTime = getAnimTrackTime(anim.animCharSet, anim.clip);
			local stopClip = false;
			
			if anim.loop then
				if anim.isActivated then
					if anim.accelerationTime ~= 0 then
						if anim.currentAcc < 1 then
							anim.currentAcc = anim.currentAcc + dt * anim.accelerationTime;
							setAnimTrackSpeedScale(anim.animCharSet, anim.clip, anim.currentAcc);
						end;
					end;
				else
					if anim.decelerationTime ~= 0 then
						if anim.currentAcc > 0 then
							anim.currentAcc = anim.currentAcc - dt * anim.decelerationTime;
							setAnimTrackSpeedScale(anim.animCharSet, anim.clip, anim.currentAcc);
						else
							stopClip = true;
						end;
					else
						stopClip = true;
					end;
				end;
			else
				if currentTime > anim.duration then
					setAnimTrackTime(anim.animCharSet, anim.clip, anim.duration);
				elseif currentTime < 0 then
					setAnimTrackTime(anim.animCharSet, anim.clip, 0);
				end;
				
				if anim.isActivated then
					if currentTime > anim.duration then
						stopClip = true;
					end;
				else
					if currentTime < 0 then
						stopClip = true;
					end;
				end;
			end;
			
			if stopClip then
				disableAnimTrack(anim.animCharSet, anim.clip);
				anim.isRunning = false;
				
				if anim.soundFile ~= nil then
					setVisibility(anim.soundFile, false);
				end;
			end;
		end;
	end;
end;

function HardPointPlaceable:setAnimationClip(id, state, noEventSend)
	local anim = self.clipAnimations[id];
	
	if anim ~= nil then
		anim.isActivated = state;
		
		if not anim.loop then
			local clipSpeed = anim.clipSpeed;
			if not state then
				clipSpeed = -clipSpeed;
			end;
			
			setAnimTrackSpeedScale(anim.animCharSet, anim.clip, clipSpeed);
			enableAnimTrack(anim.animCharSet, anim.clip);
			anim.isRunning = true;
		else
			if state then
				if anim.accelerationTime == 0 then
					anim.currentAcc = 1;
					setAnimTrackSpeedScale(anim.animCharSet, anim.clip, anim.currentAcc);
				end;
			else
				if anim.decelerationTime == 0 then
					anim.currentAcc = 0;
					disableAnimTrack(anim.animCharSet, anim.clip);
					anim.isRunning = false;
				end;
			end;
			
			anim.triggerTime = getAnimTrackTime(anim.animCharSet, anim.clip);
			enableAnimTrack(anim.animCharSet, anim.clip);
			anim.isRunning = true;
		end;
		
		if anim.soundFile ~= nil then
			setVisibility(anim.soundFile, anim.isRunning);
		end;
		
		SetAnimationEvent.sendEvent(self, id, state, noEventSend);
	end;
end;

function HardPointPlaceable:weatherChanged()
	if g_currentMission ~= nil and g_currentMission.environment ~= nil then
		for _, node in pairs(self.lightNodes) do
			setVisibility(node, not (g_currentMission.environment.isSunOn and g_currentMission.environment.currentRain == nil));
		end;
	end;
end;

function HardPointPlaceable:getBuildingSize()
	return self.buildingSize;
end;

function HardPointPlaceable:canEnterWorkshop()
	return self.allowEnterWorkshop;
end;

function HardPointPlaceable:triggerCallback(triggerId, otherId, onEnter, onLeave, onStay)
	if self.isClient then
		if g_currentMission.player ~= nil and otherId == g_currentMission.player.rootNode then
			if (onLeave or not g_currentMission.controlPlayer) then
				self.weAreInHardPointPlayerRange = nil;
			elseif onEnter and g_currentMission.controlPlayer then
				self.weAreInHardPointPlayerRange = self;
			end;
		end;
	end;
	
	if self.isEnabled then
		local vehicle = g_currentMission.nodeToVehicle[otherId];
		
		if vehicle ~= nil and vehicle.hardPointsSupported ~= nil and vehicle.hardPointsSupported then
			if onEnter then
				if self.vehiclesInRange[vehicle] == nil then
					self.vehiclesInRange[vehicle] = true;
					self.numVehicles = self.numVehicles + 1;
					vehicle:inHardPointTrigger(true);
				end;
			elseif onLeave then
				if self.vehiclesInRange[vehicle] then
					self.vehiclesInRange[vehicle] = nil;
					self.numVehicles = math.max(self.numVehicles - 1, 0);
					vehicle:inHardPointTrigger(false);
				end;
			end;
		end;
	end;
end;

registerPlaceableType("HardPointPlaceable", HardPointPlaceable);