--
-- HardPoints
--
-- @author:    	Xentro (Marcus@Xentro.se)
-- @website:	www.Xentro.se
-- @history		v1.0   - 2015-06-14 - Initial implementation
--				v1.1   - 2015-09-12 - Addon updates, Save / Load Feature
--				v1.1.1 - 2015-10-03 - Reset vehicle fix
--				v1.2   - 2015-10-25 - Seperated HP objects
-- 

source(g_currentModDirectory .. "Scripts/Events/HardPointsEvent.lua");
source(g_currentModDirectory .. "Scripts/HardPointsControlUnit.lua");

HardPoints = {};

function HardPoints.prerequisitesPresent(specializations)
    return true;
end;

function HardPoints.log(level, txt)
	if g_currentMission.HardPoint ~= nil then
		g_currentMission.HardPoint:log(level, txt);
	else
		if level == 1 then
			print("HardPoints - Error: " .. txt);
		end;
	end;
end;

function HardPoints:load(xmlFile)
	self.createObject = SpecializationUtil.callSpecializationsFunction("createObject");
	self.deleteObject = SpecializationUtil.callSpecializationsFunction("deleteObject");
	self.inHardPointTrigger = SpecializationUtil.callSpecializationsFunction("inHardPointTrigger");
	
	self.addNewMass = HardPoints.addNewMass;
	self.removeNewMass = HardPoints.removeNewMass;
	self.hardPointMass = 0;
	
	self.hardPointsSupported = true;
	self.inHardPointTriggerBool = false;
	
	self.hardPoints = {};
	self.hardPointObjects = {};
	self.activeHardpoints = {};
	
	if g_currentMission.HardPoint ~= nil then
		g_currentMission.HardPoint:loadHardPointLocationsXML(self.baseDirectory, xmlFile, self, nil, self.configFileName);
		g_currentMission.HardPoint:loadHardPointObjectsXML(self.baseDirectory, xmlFile, self, nil, self.configFileName);
	end;
	
	if self.isServer then -- we dont want clients to start doing this on join!
		local i = 0;
		while true do
			local key = string.format("vehicle.HardPoints.attachOnLoad(%d)", i);
			if not hasXMLProperty(xmlFile, key) then break; end;
			
			local locationId = getXMLInt(xmlFile, key .. "#locationId");
			local objectId = getXMLInt(xmlFile, key .. "#objectId");
			
			if locationId ~= nil and objectId ~= nil then
				if self.hardPoints[locationId] ~= nil and self.hardPointObjects[objectId] ~= nil then
					HardPoints.log(4, "ATTACH_ON_LOAD: Object ".. objectId .. " is being attached to " .. locationId);
					self:createObject(locationId, objectId);
				else
					HardPoints.log(2, "ATTACH_ON_LOAD: Invalid locationId " .. locationId .. " or objectId ".. objectId .. " for " .. key .. " in " .. self.configFileName);
				end;
			end;
			
			i = i + 1;
		end;
	end;
	
	self.ActiveObjectsDirtyFlag = self:getNextDirtyFlag();
end;

function HardPoints:delete()
	for locationId in pairs(self.activeHardpoints) do
		if g_currentMission.HardPoint ~= nil then
			-- bought
			g_currentMission.HardPoint:updateList(2, self.configFileName, self.hardPoints[locationId].attachedObjectID, -1);
		end;
		
		self:deleteObject(locationId, true);
	end;
end;

function HardPoints:loadFromAttributesAndNodes(xmlFile, key, resetVehicles)
	self.resetVehiclesHP = resetVehicles;
	
	local i = 0;
	while true do
		local hardKey = key .. string.format(".hardPoint(%d)", i);
		if not hasXMLProperty(xmlFile, hardKey) then break; end;
		
		local locationId = getXMLInt(xmlFile, hardKey .. "#locationId");
		local objectId = getXMLInt(xmlFile, hardKey .. "#objectId");
		
		if locationId ~= nil and objectId ~= nil then
			if self.loadSavedData == nil then
				self.loadSavedData = {};
			end;
			
			table.insert(self.loadSavedData, {location = locationId, object = objectId});
		end;
		
		i = i + 1;
	end;
	
	if hasXMLProperty(xmlFile, key .. ".hardPointObject(0)#locationId") then -- an object has saved data
		self.loadSavedObjectData = getXMLInt(xmlFile, key .. "#id") - 1; -- get vehicle
	end;
	
	return BaseMission.VEHICLE_LOAD_OK;
end;

function HardPoints:getSaveAttributesAndNodes(nodeIdent)
	local nodes = "";
	
	for k, v in ipairs(self.hardPoints) do
		if nodes ~= "" then
			nodes = nodes .. "\n";
		end;
		
		nodes = nodes .. nodeIdent .. '<hardPoint locationId="' .. k .. '" objectId="' .. Utils.getNoNil(v.attachedObjectID, 0) .. '" />';
	end;
	
	local attributesN, nodess;
	
	for k, v in pairs(self.activeHardpoints) do
		local attributes, subNodes = v:getSaveAttributesAndNodes(nodeIdent);
		
		if attributes ~= nil or subNodes ~= nil then
			if attributesN == nil then
				attributesN = "\n" .. nodeIdent .. '<hardPointObject locationId="' .. k .. '"';
				nodess = "";
			end;
		end;
		
		if attributes ~= nil and attributes ~= "" then
			attributesN = attributesN .. " " .. attributes;
		end;
			
		if subNodes ~= nil and subNodes ~= "" then			
			nodess = nodess .. subNodes .. "\n";
		end;
	end;
	
	if attributesN ~= nil or nodess ~= nil then
		if nodess ~= "" then
			attributesN = attributesN .. ">\n" .. nodess .. nodeIdent .. "</hardPointObject>";
		else
			attributesN = attributesN .. "/>";
		end;
		
		nodes = nodes .. attributesN;
	end;
	
	return nil, nodes;
end;

function HardPoints:readStream(streamId, connection)
	local numLocations = streamReadInt32(streamId);
	
	for locationId = 1, numLocations do
		local update = streamReadBool(streamId);
		
		if update then
			local locationId = streamReadInt8(streamId);
			local objectId = streamReadInt8(streamId);
			local serverObjectId = streamReadInt32(streamId);
			
			if self.loadSavedData == nil then
				self.loadSavedData = {};
			end;
			
			HardPoints.log(4, "READ_STREAM: Object " .. objectId .. " Location " .. locationId .. " will be attached once addon's been initialized.");
			table.insert(self.loadSavedData, {location = locationId, object = objectId, serverObjectId = serverObjectId});
		end;
	end;
end;

function HardPoints:writeStream(streamId, connection)
	streamWriteInt32(streamId, table.getn(self.hardPoints));
	
	for locationId, v in ipairs(self.hardPoints) do
		local sendUpdate = v.attachedObjectID ~= nil;
		
		streamWriteBool(streamId, sendUpdate);
		
		if sendUpdate then
			streamWriteInt8(streamId, locationId);
			streamWriteInt8(streamId, v.attachedObjectID);
			streamWriteInt32(streamId, self.activeHardpoints[locationId].id);
		end;
	end;
end;

function HardPoints:readUpdateStream(streamId, timestamp, connection)
	local doFirstSynch = streamReadBool(streamId);
	local clearDirtyFlag = streamReadBool(streamId);
	
	if doFirstSynch then
		if connection:getIsServer() then -- client
			if not self.firstObjectLoadSynch then
				HardPoints.log(4, "4. reading data from server");
				
				if streamReadInt8(streamId) ~= table.getn(self.activeHardpoints) then
					HardPoints.log(1, "MP stream is out of synch!");
				end;
				
				for locationId, v in pairs(self.activeHardpoints) do
					HardPoints.log(4, "-- " .. locationId .. " reading locationId");
					v:readStream(streamId, connection);
				end;
				
				if clearDirtyFlag then
					self.firstObjectLoadSynch = doFirstSynch;
					self:clearDirtyFlags(self.ActiveObjectsDirtyFlag); 
				end;
			end;
			
		else -- server 
			HardPoints.log(4, "2. server writing dirty update to clients");
			
			self:raiseDirtyFlags(self.ActiveObjectsDirtyFlag);
		end;
	else
		if not connection:getIsServer() then -- server
			if clearDirtyFlag and self.initializeStartUp then
				self:clearDirtyFlags(self.ActiveObjectsDirtyFlag);
			end;
		end;
	end;
	
	if not doFirstSynch then
		-- we are finished, object will sort rest.
	end;
end;

function HardPoints:writeUpdateStream(streamId, connection, dirtyMask)
	if not connection:getIsServer() then -- server
		local doFirstSynch = bitAND(dirtyMask, self.ActiveObjectsDirtyFlag) ~= 0;
		
		streamWriteBool(streamId, doFirstSynch);
		streamWriteBool(streamId, true); -- clearDirtyFlag
		
		if doFirstSynch then
			HardPoints.log(4, "3. writing data to clients");
			
			streamWriteInt8(streamId, table.getn(self.activeHardpoints));
			
			for locationId, v in pairs(self.activeHardpoints) do
				HardPoints.log(4, "-- " .. locationId .. " writing locationId");
				v:writeStream(streamId, connection);
			end;
		end;
		
	else -- client
		if not self.firstObjectLoadSynch and self.initializeStartUp then
			HardPoints.log(4, "1. client writing false update");
			
			streamWriteBool(streamId, true); -- doFirstSynch
			streamWriteBool(streamId, false); -- clearDirtyFlag
		else
			streamWriteBool(streamId, false); -- doFirstSynch
			streamWriteBool(streamId, true); -- clearDirtyFlag
		end;
	end;
end;

function HardPoints:mouseEvent(posX, posY, isDown, isUp, button)
end;

function HardPoints:keyEvent(unicode, sym, modifier, isDown)
end;

function HardPoints:update(dt)
	if self.firstTimeRun and not self.initializeStartUp then -- addons are loaded on the first run
		if g_currentMission.HardPoint ~= nil then
			local addon = g_currentMission.HardPoint.addonsToLoad[self.configFileName];
			
			if addon ~= nil then
				HardPoints.log(3, "Starting to load addon data");
				
				for _, a in ipairs(addon) do
					local addonXmlFile = loadXMLFile("TempAddonObject", a.fileName);

					g_currentMission.HardPoint:loadHardPointLocationsXML(a.baseDirectory, addonXmlFile, self, table.getn(self.hardPoints), a.fileName); -- something is bugged in here
					g_currentMission.HardPoint:loadHardPointObjectsXML(a.baseDirectory, addonXmlFile, self, table.getn(self.hardPointObjects), a.fileName, a.customEnvironment);
					
					HardPoints.log(4, "LOAD_ADDON: Addons from " .. a.fileName .. " have been loaded.");
					
					delete(addonXmlFile);
				end;
			end;
		end;
	
		if self.loadSavedData ~= nil then
			HardPoints.log(3, "Starting to load either saved data or sent data from server.");
			
			for _, v in ipairs(self.loadSavedData) do
				if self.hardPoints[v.location] ~= nil then
					local location = self.hardPoints[v.location];
					if v.object == 0 then
						if location.attachedObjectID ~= nil then
							HardPoints.log(2, "LOADING_SAVE_OR_PLAYER_JOINED: ObjectID is 0 but we got something attached, emptying location ( " .. v.location .. " ), corrupt save?");
							self:deleteObject(v.location, true);
						end;
					else
						if (location.attachedObjectID == nil or location.attachedObjectID ~= nil and location.attachedObjectID ~= v.object) then
							if self.hardPointObjects[v.object] ~= nil then
								self:createObject(v.location, v.object, true, v.serverObjectId);
							else
								HardPoints.log(2, "LOADING_SAVE_OR_PLAYER_JOINED: Is trying to load object id ( " .. v.object .. " ) but its invalid, possible reason could be corrupt save or that you removed an addon mod that this mod use.");
							end;
						end;
					end;
				else
					HardPoints.log(2, "LOADING_SAVE_OR_PLAYER_JOINED: Is trying to load location ( " .. v.location .. " ) but its invalid, possible reason could be corrupt save or that you removed an addon mod that this mod use.");
				end;
			end;
			
			self.loadSavedData = nil;
		end;
		
		-- load saved game again for object data
		if self.loadSavedObjectData ~= nil then
			if g_currentMission.missionInfo.vehiclesXMLLoad ~= nil then
				HardPoints.log(3, "We got saved object data to load.");
				
				local xmlFile = loadXMLFile("VehiclesXML", g_currentMission.missionInfo.vehiclesXMLLoad);
				
				local i = 0;
				while true do
					local key = string.format("careerVehicles.vehicle(%d).hardpointObject(%d)", self.loadSavedObjectData, i);
					if not hasXMLProperty(xmlFile, key) then break; end;
					
					local locationId = getXMLInt(xmlFile, key .. "#locationId");
					if locationId ~= nil then
						if self.activeHardpoints[locationId] ~= nil then					
							self.activeHardpoints[locationId]:loadAttributesAndNodes(xmlFile, key, Utils.getNoNil(self.resetVehiclesHP, false));
						else
							HardPoints.log(2, "LOADING_OBJECT_SAVE: An object failed to load its saved data, possible reason is that location ( " .. locationId .. " ) are an addon.");
						end;
					end;
					
					i = i + 1;
				end;
				
				delete(xmlFile);
			end;
			
			self.loadSavedObjectData = nil;
			self.resetVehiclesHP = nil;
		end;
		
		for locationId in ipairs(self.hardPoints) do
			HardPoints.updateChildLocations(self, locationId);
		end;
		
		if self.isClient then
			self:raiseDirtyFlags(self.ActiveObjectsDirtyFlag); -- synch object data as player joined
		end;
		
		HardPoints.log(3, "HardPoint have been Initialized for " ..  StoreItemsUtil.storeItemsByXMLFilename[self.configFileName:lower()].name);
		self.initializeStartUp = true;
	end;
end;

function HardPoints:updateTick(dt)
end;

function HardPoints:draw()
	if self.inHardPointTriggerBool and self:getIsActive() then
		g_currentMission:addExtraPrintText(string.format(g_i18n:getText("HARDPOINT_NOTICE_5"), StoreItemsUtil.storeItemsByXMLFilename[self.configFileName:lower()].name));
	end;
	
	for _, v in pairs(self.activeHardpoints) do
		v:draw();
	end;
end;

function HardPoints:createObject(locationId, objectId, forceCreate, serverId)
	if forceCreate == nil then
		forceCreate = false;
	end;
	
	local hp = self.hardPoints[locationId];
	local object = self.hardPointObjects[objectId];
	
	if hp ~= nil and object ~= nil then
		if hp.attachedObjectID ~= nil and hp.attachedObjectID ~= objectId then
			HardPoints.log(4, "ATTACHING: Force delete on location '" .. hp.type .. "' id: " .. locationId .. ", object " .. object.name .. " id: " .. objectId );
			self:deleteObject(locationId);
		end;
		
		if hp.attachedObjectID == nil or hp.attachedObjectID ~= nil and forceCreate then
			hp.attachedObjectID = objectId; -- pre set for calling client
			
			if self.initializeStartUp then
				HardPoints.updateChildLocations(self, locationId);
			end;
			
			if not self.isServer and not forceCreate then
				HardPointsEvent.sendEvent(self, HardPointsEvent.CREATE_OBJECT_SERVER, false, locationId, objectId); -- tell server to create object
			else
				local hpcu = HardPointsControlUnit:new(self.isServer, self.isClient);
				local isOkay = hpcu:load(self, hp.nodes, locationId, object);
				
				if isOkay then
					self.activeHardpoints[locationId] = hpcu;
					hpcu:register(true);
					
					if not self.isServer then
						g_client:finishRegisterObject(hpcu, serverId);
						HardPointsEvent.sendEvent(self, HardPointsEvent.FINSIH_SERVER, false, nil, nil, hpcu); -- client are creating object now finish this
					end;
					
					HardPointsEvent.sendEvent(self, HardPointsEvent.CREATE_OBJECT_CLIENT, not self.isServer, locationId, objectId, hpcu.id); -- server will send an event to tell client to create object, clients wont send even
					HardPoints.log(3, "ATTACHING: " .. object.name .. ", " .. objectId .. " have been attached to location '" .. hp.type .. "' id: " .. locationId);
				else
					-- we assume this wont happen, seeing as you setup the object as it should!
					hpcu:delete();
					self:deleteObject(locationId);
					HardPoints.log(1, "ATTACHING: " .. object.name .. ", " .. objectId .. " failed to attach to location '" .. hp.type .. "' id: " .. locationId);
				end;
			end;
		else
			HardPoints.log(4, "ATTACHING: " .. self.hardPointObjects[hp.attachedObjectID].name .. ", " .. hp.attachedObjectID .. " attached already on location '" .. hp.type .. "' id:" .. locationId);
		end;
	else
		HardPoints.log(2, "ATTACHING: Invalid location (" .. tostring(locationId) .. ") or objectId (" .. tostring(objectId) .. ")");
	end;
end;

function HardPoints:deleteObject(locationId, noEventSend)
	local hp = self.hardPoints[locationId];
	
	if hp ~= nil then
		if hp.attachedObjectID ~= nil then
			hp.attachedObjectID = nil;
			
			if self.activeHardpoints[locationId] ~= nil then
				self.activeHardpoints[locationId]:delete();
				self.activeHardpoints[locationId] = nil;
				HardPoints.log(4, "DETACHING: Deleting active object");
			end;
			
			HardPoints.updateChildLocations(self, locationId);
			HardPoints.log(3, "DETACHING: Object on location '" .. hp.type .. "' id:" .. locationId .. " have been deleted.");
		end;
		
		HardPointsEvent.sendEvent(self, HardPointsEvent.DELETE_OBJECT, noEventSend, locationId);
	else
		HardPoints.log(2, "DETACHING: Invalid location " .. tostring(locationId));
	end;
end;

function HardPoints:updateChildLocations(locationId)
	local location = self.hardPoints[locationId];
	local hasObjectAttached = location.attachedObjectID ~= nil;
	
	if location.childIds ~= nil then
		for _, childId in pairs(location.childIds) do
			local childLocation = self.hardPoints[childId];
			
			if childLocation ~= nil then
				local childHasObjectAttached = childLocation.attachedObjectID ~= nil;
				childLocation.weAreChild = true;
				childLocation.parentAllowAdding = hasObjectAttached;
				
				if childLocation.parentAllowAdding then
					childLocation.parentDisabledId = nil;
				else
					childLocation.parentDisabledId = locationId;
				end;
				
				HardPoints.log(4, "UPDATE_CHILD: parentAllowAdding:" .. tostring(childLocation.parentAllowAdding) .. ", for location " .. childLocation.type .. " " .. childId);
				
				if childHasObjectAttached and not childLocation.parentAllowAdding then
					g_currentMission.HardPoint:updateList(1, self.configFileName, childLocation.attachedObjectID, 1);
					
					self:deleteObject(childId);
					HardPoints.log(4, "UPDATE_CHILD: Parent location has been emptied, detaching child location " .. childLocation.type .. " " .. childId);
				end;
			end;
		end;
	end;
end;


-- mass section
function HardPoints:addNewMass(mass)
	self.hardPointMass = self.hardPointMass + mass;
	HardPoints.log(4, "MASS: adding mass (" .. mass .. "), new total is " .. self.hardPointMass);
end;

function HardPoints:removeNewMass(mass)
	self.hardPointMass = math.max(self.hardPointMass - mass, 0);
	HardPoints.log(4, "MASS: removing mass (" .. mass .. "), new total is " .. self.hardPointMass);
end;

if not Utils.getTotalMassFix20150914 then
	local getTotalMass = Vehicle.getTotalMass;
	Vehicle.getTotalMass = function(self, addAttachables, addOnlyNonWheelAttachables)
		local mass = getTotalMass(self, addAttachables, addOnlyNonWheelAttachables);
		
		if self.isServer and self.hardPointsSupported ~= nil and self.hardPointsSupported then
			mass = mass + self.hardPointMass;
		end;
		
		return mass;
	end;
	
	Utils.getTotalMassFix20150914 = true;
end;


-- in range message for vehicle
function HardPoints:inHardPointTrigger(bool)
	self.inHardPointTriggerBool = bool;
end;


-- reload for objects
function HardPoints:developmentReloadFromXML(xmlFile)
	for _, v in pairs(self.activeHardpoints) do
		v:developmentReloadFromXML();
	end;
end;

-- TODO: to be used when changing location stuff ingame, shitload of extra work!
-- function HardPoints:addExtraLocation(location)
-- function HardPoints:removeExtraLocation(locationId)