--
-- HardPointsControlUnit
--
-- @author:    	Xentro (Marcus@Xentro.se)
-- @website:	www.Xentro.se
-- @history		v1.0 - 2015-06-13 - Initial implementation
-- 				v1.1 - 2015-09-20 - Save / Load Feature, Color selection synch, MP Synch functions
--				v1.2 - 2015-10-03 - New tags for collision and some new functions
--				v1.3 - 2015-10-26 - Object is its own object class (MP thing)
-- 

source(g_currentModDirectory .. "Scripts/Events/SetInActiveStateEvent.lua");

HardPointsControlUnit = {};
HardPointsControlUnit_mt = Class(HardPointsControlUnit, Object);

InitObjectClass(HardPointsControlUnit, "HardPointsControlUnit");

HardPointsControlUnit.LOAD_OK = true;
HardPointsControlUnit.LOAD_FAIL = false;

function HardPointsControlUnit:new(isServer, isClient, customMt)
	HardPoints.log(3, "CONTROL_UNIT: Starting");
	
	local mt = customMt;
	if mt == nil then
		mt = HardPointsControlUnit_mt;
	end;
	
    local instance = Object:new(isServer, isClient, mt);
	
	instance.specializations = {};
	instance.components = {};
	
	instance.weAreHardPointObject = true;
	instance.controlUnitVersion = 1.33;
	instance.firstTimeRun = false;
	
	return instance;
end;

function HardPointsControlUnit:load(vehicle, locationNodes, locationId, object)
	self.object = object;
	self.locationId = locationId;
	
	self.vehicle = vehicle;
	self.environment = self;
	self.specializations = self.object.specs;
	self.configFileName = self.object.xmlFilename;
	
	self.componentJointsData = {};
	self.dirtNodes = {};
    self.skinnedNodes = {};
	
	self.inActiveState = true; -- deactivate script updates
	self.forceIsActive = false;
	self.baseDirectory = self.vehicle.baseDirectory;
	self.firstCreateStreamDone = false;
	
	local xmlFile;
	local failedToLoad = false;
	local tempRootNode = createTransformGroup("tempRootNode");
	local Root = Utils.loadSharedI3DFile(self.object.filename);
	local noneColI, colI = 0, 0;
	
	for i in ipairs(locationNodes) do
		local cNode = getChildAt(Root, 0);

		if cNode ~= nil then
			HardPoints.log(4, "CONTROL_UNIT: " .. getName(cNode) .. " component is loaded.");
			
			if self.rootNode == nil then
				self.rootNode = cNode;
			end;
			
			local component = {};
			component.node = cNode;
			component.hasCollision = getRigidBodyType(cNode) == "Dynamic";
			
			if component.hasCollision then
				if not self.isServer then
					setRigidBodyType(component.node, "Kinematic");
				end;
				
				colI = colI + 1;
			else
				if not getVisibility(component.node) then
					HardPoints.log(2, "CONTROL_UNIT: " .. getName(component.node) .. " will not be shown ingame as its visibility is unchecked");
				end;
				
				noneColI = noneColI + 1;
			end;
			
			link(tempRootNode, component.node);
			translate(component.node, 0, 0, 0);
			
			table.insert(self.components, component);
		else
			failedToLoad = true;
			break;
		end;
	end;

	delete(Root);
	
	local numLoc, numComp = table.getn(locationNodes), table.getn(self.components);
	if numLoc ~= numComp then
		HardPoints.log(1, "CONTROL_UNIT: " .. self.object.name .. " don't have the same number of components (" .. numComp .. ") as location nodes (" .. numLoc .. ").");
		failedToLoad = true;
	end;
	
	if not HardPointsControlUnit.canKeepLoading(failedToLoad, tempRootNode) then
		return HardPointsControlUnit.LOAD_FAIL;
	end;
	
	if self.configFileName ~= nil then
		xmlFile = loadXMLFile("TempConfig", self.configFileName);
	end;
	
	if xmlFile ~= nil then
		local i = 0;
		while true do
			local key = string.format("vehicle.components.component(%d)", i);
			if not hasXMLProperty(xmlFile, key) then break; end;
			
			local id = getXMLInt(xmlFile, key .. "#id");
			
			local offsetTrans = Utils.getVectorNFromString(getXMLString(xmlFile, key .. "#offsetTrans"), 3);
			local offsetRot = Utils.getRadiansFromString(getXMLString(xmlFile, key .. "#offsetRot"), 3);
			
			if id ~= nil then 
				local component = self.components[id + 1];
				
				if component ~= nil then
					component.offsetTrans = offsetTrans;
					component.offsetRot = offsetRot;
					
					if self.isServer then
						local mass = getXMLFloat(xmlFile, key .. "#mass");
			
						if mass ~= nil then
							if component.hasCollision then
								setMass(component.node, mass);
							end;
							
							component.mass = mass;
							self.vehicle:addNewMass(mass);
						else
							if component.hasCollision then
								-- issue, mass is reset to 1 after object is created the second time
								component.mass = getMass(component.node);
								self.vehicle:addNewMass(component.mass);
								
								HardPoints.log(2, "CONTROL_UNIT - BUG: Currently there is an issue with mass of objects, Work around is to set the mass in xml. Object: " .. self.object.name .. " and component id " .. id + 1);
								HardPoints.log(2, "CONTROL_UNIT - BUG: Object Mass: " .. component.mass); -- remove
							end;
						end;
					end;
					
					if component.hasCollision then
						local jointIndex = getXMLString(xmlFile, key .. "#jointIndex");
						local vId = getXMLInt(xmlFile, key .. "#vehicleComponentId");
						local cId = getXMLInt(xmlFile, key .. "#targetComponentId");
					
						if jointIndex ~= nil and (vId ~= nil or cId ~= nil) then
							if (vId ~= nil and self.vehicle.components[vId + 1] ~= nil or cId ~= nil and self.components[cId + 1] ~= nil) then
								local jointNode = Utils.indexToObject(self.components, jointIndex);
								
								if jointNode ~= nil then
									component.jointNode = jointNode;
									component.jointNodeActor1 = Utils.getNoNil(Utils.indexToObject(self.components, getXMLString(xmlFile, key .. "#indexActor1")), component.jointNode);
									
									if vId ~= nil then
										component.vId = vId + 1;
									else
										component.cId = cId + 1;
									end;
									
									local x, y, z = Utils.getVectorFromString(getXMLString(xmlFile, key .. "#centerOfMass"));
									if x ~= nil and y ~= nil and z ~= nil then
										setCenterOfMass(component.node, x, y, z);
										component.centerOfMass = {x, y, z};
									end;
									
									local count = getXMLInt(xmlFile, key .. "#solverIterationCount");
									if count ~= nil then
										setSolverIterationCount(component.node, count);
										component.solverIterationCount = count;
									end;
									
									local x, y, z = Utils.getVectorFromString(getXMLString(xmlFile,  key .. "#rotLimit"));
									local rotLimits = {math.rad(Utils.getNoNil(x, 0)), math.rad(Utils.getNoNil(y, 0)), math.rad(Utils.getNoNil(z, 0))};

									local x, y, z = Utils.getVectorFromString(getXMLString(xmlFile,  key .. "#transLimit"));
									local transLimits = {Utils.getNoNil(x, 0), Utils.getNoNil(y, 0), Utils.getNoNil(z, 0)};

									component.rotLimit = rotLimits;
									component.transLimit = transLimits;
									
									local x, y, z = Utils.getVectorFromString(getXMLString(xmlFile,  key .. "#rotMinLimit"));
									local rotMinLimits = {Utils.getNoNilRad(x, nil), Utils.getNoNilRad(y, nil), Utils.getNoNilRad(z, nil)};

									local x, y, z = Utils.getVectorFromString(getXMLString(xmlFile,  key .. "#transMinLimit"));
									local transMinLimits = {x, y, z};

									for i = 1,3 do
										if rotMinLimits[i] == nil then
											if rotLimits[i] >= 0 then
												rotMinLimits[i] = -rotLimits[i];
											else
												rotMinLimits[i] = rotLimits[i] + 1;
											end;
										end;
										if transMinLimits[i] == nil then
											if transLimits[i] >= 0 then
												transMinLimits[i] = -transLimits[i];
											else
												transMinLimits[i] = transLimits[i] + 1;
											end;
										end;
									end;

									component.rotMinLimit = rotMinLimits;
									component.transMinLimit = transMinLimits;
					
									local x, y, z = Utils.getVectorFromString(getXMLString(xmlFile,  key .. "#rotLimitSpring"));
									local rotLimitSpring = {Utils.getNoNil(x, 0), Utils.getNoNil(y, 0), Utils.getNoNil(z, 0)};
									local x, y, z = Utils.getVectorFromString(getXMLString(xmlFile,  key .. "#rotLimitDamping"));
									local rotLimitDamping = {Utils.getNoNil(x, 1), Utils.getNoNil(y, 1), Utils.getNoNil(z, 1)};
									component.rotLimitSpring = rotLimitSpring;
									component.rotLimitDamping = rotLimitDamping;
									
									local x, y, z = Utils.getVectorFromString(getXMLString(xmlFile,  key .. "#transLimitSpring"));
									local transLimitSpring = {Utils.getNoNil(x, 0), Utils.getNoNil(y, 0), Utils.getNoNil(z, 0)};
									local x, y, z = Utils.getVectorFromString(getXMLString(xmlFile,  key .. "#transLimitDamping"));
									local transLimitDamping = {Utils.getNoNil(x, 1), Utils.getNoNil(y, 1), Utils.getNoNil(z, 1)};
									component.transLimitSpring = transLimitSpring;
									component.transLimitDamping = transLimitDamping;
									
									i = i + 1;
								else
									HardPoints.log(1, "CONTROL_UNIT: JointNode is nil for " .. self.object.name .. ", jointNode is used to attach collision to vehicle.");
									failedToLoad = true;
									break;
								end;
							else
								if vId ~= nil then
									HardPoints.log(1, "CONTROL_UNIT: Invalid vehicleComponentId " .. vId + 1);
								else
									HardPoints.log(1, "CONTROL_UNIT: Invalid targetComponentId " .. cId + 1);
								end;
								
								failedToLoad = true;
								break;
							end;
						else
							HardPoints.log(2, "CONTROL_UNIT: Object " ..  self.object.name .. " have collision but no jointIndex or vehicleComponentId or targetComponentId, result will be collision won't be attached.");
							i = i + 1;
						end;
					else
						i = i + 1;
					end;
				else
					HardPoints.log(1, "CONTROL_UNIT: Invalid component id " .. id + 1 .. ".");
					failedToLoad = true;
					break;
				end;
			else
				i = i + 1;
			end;
		end;
		
		local color = self.vehicle.color;
		if color ~= nil and Utils.getNoNil(getXMLBool(xmlFile, "vehicle.colorNodes#changeColor"), true) then
			local i = 0;
			while true do
				local key = string.format("vehicle.colorNodes.colorNode(%d)", i);
				if not hasXMLProperty(xmlFile, key) then break; end;
				
				local node = Utils.indexToObject(self.components, getXMLString(xmlFile, key .. "#index"));
				if node ~= nil then
					if self.colorNodes == nil then
						self.colorNodes = {};
					end;
					
					table.insert(self.colorNodes, node);
				end;
				
				i = i + 1;
			end;
			
			self:setColor(color[1], color[2], color[3], color[4]); -- use default function but set for object.
			HardPoints.log(4, "CONTROL_UNIT: Changing color of object.");
		end;
		
		local i = 0;
		while true do
			local key = string.format("vehicle.skinnedNodes.skinnedNode(%d)", i);
			if not hasXMLProperty(xmlFile, key) then break; end;
			
			local node = Utils.indexToObject(self.components, getXMLString(xmlFile, key .. "#index"));
			if node ~= nil then
				if Utils.getNoNil(getXMLBool(xmlFile, key .. "#moveToZero"), false) then
					setTranslation(node, 0,0,0);
					setRotation(node, 0,0,0);
				end;
				
				table.insert(self.skinnedNodes, node);
			end;
			i = i + 1;
		end;
		
		if g_currentMission.paintWorkshop ~= nil then -- paint mod is present
			self.materialNodes = {};
			
			if self.vehicle.loadMaterialsXML ~= nil then -- parent vehicle supports paint mod
				self.vehicle.loadMaterialsXML(self, xmlFile, self.vehicle.configFileName);
				self.vehicle.loadPaintNodesXML(self, xmlFile);
			end;
		end;
	end;
	
	if not HardPointsControlUnit.canKeepLoading(failedToLoad, tempRootNode, xmlFile) then
		return HardPointsControlUnit.LOAD_FAIL;
	end;
	
	for i, v in ipairs(locationNodes) do
		local component = self.components[i];
		
		Utils.getNodesByShaderParam(component.node, "RDT", self.dirtNodes); -- get dirt nodes
		
		if not component.hasCollision then			
			link(v.node, component.node);
			
			if component.offsetTrans ~= nil then
				setTranslation(component.node, unpack(component.offsetTrans));
			end;
			if component.offsetRot ~= nil then
				setRotation(component.node, unpack(component.offsetRot));
			end;
		else
			if component.jointNode ~= nil then
				if self.isServer then
					simulatePhysics(false);
				end;
				
				local node;
				
				if component.cId == nil then
					node = v.node;
				else
					node = component.jointNode;
				end;
				
				local trans = {getWorldTranslation(node)};
				local rot = {getWorldRotation(node)};
				
				if component.offsetTrans ~= nil then
					for Io, Ot in ipairs(component.offsetTrans) do
						trans[Io] = trans[Io] + Ot;
					end;
				end;
				if component.offsetRot ~= nil then
					for Io, Ot in ipairs(component.offsetRot) do
						rot[Io] = rot[Io] + Ot;
					end;
				end;
				
				setTranslation(tempRootNode, unpack(trans));
				setRotation(tempRootNode, unpack(rot));
				
				setTranslation(component.node, unpack(trans));
				setRotation(component.node, unpack(rot));
				link(getRootNode(), component.node);
				
				if self.isServer then
					local constr = JointConstructor:new();
					if component.vId ~= nil then
						constr:setActors(self.vehicle.components[component.vId].node, component.node);
					else
						constr:setActors(self.components[component.cId].node, component.node);
					end;
					
					constr:setJointTransforms(component.jointNode, component.jointNodeActor1);
					constr:setRotationLimitSpring(component.rotLimitSpring[1], component.rotLimitDamping[1], component.rotLimitSpring[2], component.rotLimitDamping[2], component.rotLimitSpring[3], component.rotLimitDamping[3]);
					constr:setTranslationLimitSpring(component.transLimitSpring[1], component.transLimitDamping[1], component.transLimitSpring[2], component.transLimitDamping[2], component.transLimitSpring[3], component.transLimitDamping[3]);

					for i = 1, 3 do
						if component.rotLimit[i] >= component.rotMinLimit[i] then
							constr:setRotationLimit(i - 1, component.rotMinLimit[i], component.rotLimit[i]);
						end;

						if component.transLimit[i] >= component.transMinLimit[i] then
							constr:setTranslationLimit(i - 1, true, component.transMinLimit[i], component.transLimit[i]);
						else
							constr:setTranslationLimit(i - 1, false, 0, 0);
						end;
					end;
					
					table.insert(self.componentJointsData, {componentId = i, jointIndex = constr:finalize(), jointNode = node});
					simulatePhysics(true);
				end;
			else
				-- reason have been told already, lets keep it going.
			end;
		end;
	end;
	
	if not HardPointsControlUnit.canKeepLoading(failedToLoad, tempRootNode, xmlFile) then
		return HardPointsControlUnit.LOAD_FAIL;
	end;
	
	if xmlFile ~= nil then
		local i = 0;
		while true do
			local key = string.format("vehicle.components.collisionPair(%d)", i);
			if not hasXMLProperty(xmlFile, key) then break; end;
			
			local enabled = Utils.getNoNil(getXMLBool(xmlFile, key .. "#enabled"), false);
			local index0 = getXMLInt(xmlFile, key .. "#vehicleComponent");
			local index1 = getXMLInt(xmlFile, key .. "#component1");
			local index2 = getXMLInt(xmlFile, key .. "#component2");
			
			if (index0 ~= nil or index1 ~= nil) and index2 ~= nil then
				local component1;
				
				if index0 ~= nil then
					component1 = self.vehicle.components[index0 + 1];
				elseif index1 ~= nil then
					component1 = self.components[index1 + 1];
				end;
				
				local component2 = self.components[index2 + 1];
				
				if component1 ~= nil and component2 ~= nil then
					setPairCollision(component1.node, component2.node, enabled);
					HardPoints.log(4, "CONTROL_UNIT: Setting collision pair between component " .. getName(component1.node) .. " and " .. getName(component2.node));
				end;
			end;
			
			i = i + 1;
		end;
	end;
	
	for _, v in ipairs(self.specializations) do
		if v.preLoad ~= nil then
			v.preLoad(self.environment, xmlFile);
		end;
	end;
	
	if self.vehicle.addWashableNode ~= nil then
		for _, node in pairs(self.dirtNodes) do
			self.vehicle:addWashableNode(node);
		end;
		
		self.vehicle:setDirtAmount(self.vehicle:getDirtAmount(), true); -- object cant have its own value.
		HardPoints.log(4, "CONTROL_UNIT: Setting dirt value.");
	end;
	
	for _, node in pairs(self.skinnedNodes) do
        local minClipDistance = getClipDistance(node);
        local curNode = getParent(node);
        while curNode ~= 0 do
            minClipDistance = math.min(minClipDistance, getClipDistance(curNode));
            curNode = getParent(curNode);
        end
        setClipDistance(node, minClipDistance)

        link(getRootNode(), node);
    end;
	
	for _, v in ipairs(self.specializations) do
		if v.load ~= nil then
			v.load(self.environment, xmlFile);
		end;
	end;
	
	local weAreFinsihed = HardPointsControlUnit.canKeepLoading(true, tempRootNode, xmlFile); -- finish loading by cleaning
	return HardPointsControlUnit.LOAD_OK;
end;

function HardPointsControlUnit:delete()
	for _, v in ipairs(self.specializations) do
		if v.delete ~= nil then
			v.delete(self.environment);
		end;
	end;
	
	if self.isServer then
		for _, joint in ipairs(self.componentJointsData) do
			removeJoint(joint.jointIndex);
		end;
	end;
	
	if self.vehicle.removeWashableNode ~= nil then
		for _, node in pairs(self.dirtNodes) do
			self.vehicle:removeWashableNode(node);
		end;
	end;
	
	for i = 1, table.getn(self.components) do
		if self.isServer then
			if self.components[i].mass ~= nil then
				HardPoints.log(4, "MASS: " .. self.components[i].mass);
				self.vehicle:removeNewMass(self.components[i].mass);
			end;
		end;
		
		if self.components[i].node ~= nil and entityExists(self.components[i].node) then
			if not self.components[i].hasCollision then
				unlink(self.components[i].node);
			end;
			
			delete(self.components[i].node);
		end;
	end;
	
    for _, node in pairs(self.skinnedNodes) do
		delete(node);
    end;
	
	Utils.releaseSharedI3DFile(self.object.filename, nil, true);
	
	if self.isRegistered then
		if self.isServer then
			g_server:unregisterObject(self, true);
		else
			g_client:unregisterObject(self, true);
		end;
	end;
	
	HardPoints.log(4, "CONTROL_UNIT: Deleting Object");
end;

function HardPointsControlUnit:loadAttributesAndNodes(xmlFile, key, resetVehicles)
	for _, v in ipairs(self.specializations) do
		if v.loadAttributesAndNodes ~= nil then
			v.loadAttributesAndNodes(self.environment, xmlFile, key, resetVehicles);
		end;
	end;
end;	

function HardPointsControlUnit:getSaveAttributesAndNodes(nodeIdent)
	local attributes, nodes;
	
	for _, v in ipairs(self.specializations) do
		if v.getSaveAttributesAndNodes ~= nil then
			local specAttributes, specNodes = v.getSaveAttributesAndNodes(self.environment, nodeIdent);
			
			if specAttributes ~= nil and specAttributes ~= "" then
				if attributes == nil then
					attributes = "";
				end;
				attributes = attributes .. " " .. specAttributes;
			end;
			
			if specNodes ~= nil and specNodes ~= "" then
				if nodes == nil then
					nodes = specNodes;
				else
					nodes = nodes .. "\n" .. specNodes;
				end;
			end;
		end;
	end;
	
	return attributes, nodes;
end;

function HardPointsControlUnit:readStream(streamId, connection)
	self:setInActiveState(streamReadBool(streamId), true);
	
	for _, v in ipairs(self.specializations) do
		if v.readStream ~= nil then
			v.readStream(self.environment, streamId, connection);
		end;
	end;
end;

function HardPointsControlUnit:writeStream(streamId, connection)
	streamWriteBool(streamId, self.inActiveState);
	
	for _, v in ipairs(self.specializations) do
		if v.writeStream ~= nil then
			v.writeStream(self.environment, streamId, connection);
		end;
	end;
end;

function HardPointsControlUnit:readUpdateStream(streamId, timestamp, connection)
	for _, v in ipairs(self.specializations) do
		if v.readUpdateStream ~= nil then
			v.readUpdateStream(self.environment, streamId, timestamp, connection);
		end;
	end;
end;

function HardPointsControlUnit:writeUpdateStream(streamId, connection, dirtyMask)
	for _, v in ipairs(self.specializations) do
		if v.writeUpdateStream ~= nil then
			v.writeUpdateStream(self.environment, streamId, connection, dirtyMask);
		end;
	end;
end;

function HardPointsControlUnit:update(dt)
	for _, v in ipairs(self.specializations) do
		if v.update ~= nil and self.inActiveState then
			v.update(self.environment, dt);
		end;
	end;
	
	self.firstTimeRun = true;
end;

function HardPointsControlUnit:updateTick(dt)
	for _, v in ipairs(self.specializations) do
		if v.updateTick ~= nil and self.inActiveState then
			v.updateTick(self.environment, dt);
		end;
	end;
end;

function HardPointsControlUnit:draw()
	for _, v in ipairs(self.specializations) do
		if v.draw ~= nil and self.inActiveState then
			v.draw(self.environment);
		end;
	end;
end;


-- some joint stuff
function HardPointsControlUnit:updateJoint(jointIndex)
	if self.isServer then
		local joint = self.componentJointsData[jointIndex];
		setJointFrame(joint.jointIndex, 0, joint.jointNode);
	end;
end;

function HardPointsControlUnit:setComponentJointRotLimit(jointIndex, axis, minLimit, maxLimit)
    if self.isServer then
		local joint = self.componentJointsData[jointIndex];
        joint.rotLimit[axis] = maxLimit;
        joint.rotMinLimit[axis] = minLimit;

        if minLimit <= maxLimit then
            setJointRotationLimit(joint.jointIndex, axis - 1, true, minLimit, maxLimit);
        else
            setJointRotationLimit(joint.jointIndex, axis - 1, false, 0, 0);
        end;
    end;
end;

function HardPointsControlUnit:setComponentJointTransLimit(jointIndex, axis, minLimit, maxLimit)
    if self.isServer then
		local joint = self.componentJointsData[jointIndex];
        joint.transLimit[axis] = maxLimit;
        joint.transMinLimit[axis] = minLimit;

        if minLimit <= maxLimit then
            setJointTranslationLimit(joint.jointIndex, axis - 1, true, minLimit, maxLimit);
        else
            setJointTranslationLimit(joint.jointIndex, axis - 1, false, 0, 0);
        end;
    end;
end;


-- disable big part of the object
function HardPointsControlUnit:setInActiveState(inActiveState, noEventSend)
	self.inActiveState = inActiveState;
	
	for _, v in ipairs(self.specializations) do
		if v.setInActiveState ~= nil then
			v.setInActiveState(self.environment, inActiveState);
		end;
	end;
	
	SetInActiveStateEvent.sendEvent(self, inActiveState, noEventSend);
end;


-- direct to the vehicle for guidance 
function HardPointsControlUnit:getIsActive()
	if self.forceIsActive then
		return true;
	end;
	
	return self.vehicle:getIsActive();
end;

function HardPointsControlUnit:getIsActiveForInput(bool)
	return self.vehicle:getIsActiveForInput(bool);
end;

function HardPointsControlUnit:getIsActiveForSound()
	return self.vehicle:getIsActiveForSound();
end;

function HardPointsControlUnit:hasInputConflictWithSelection(inputs)
	return self.vehicle:hasInputConflictWithSelection(inputs);
end;

function HardPointsControlUnit:addConflictCheckedInput(actionIndex)
	return self.vehicle:addConflictCheckedInput(actionIndex);
end;

function HardPointsControlUnit:removeConflictCheckedInput(actionIndex)
	return self.vehicle:removeConflictCheckedInput(actionIndex);
end;

function HardPointsControlUnit:getOwner()
    return self.vehicle:getOwner();
end;


-- color stuff 
function HardPointsControlUnit:setColor(r, g, b, a)
	self.vehicle.setColor(self, r, g, b, a); -- use default function but set for object.
end;

function HardPointsControlUnit:setMaterial(paint, mat, id, lastId)
	if self.materialNodes ~= nil then
		if id ~= lastId then
			for _, v in pairs(paint[lastId]) do
				for _, node in pairs(self.materialNodes[v.name]) do
					setVisibility(node, false);
				end;
			end;
		end;
		
		for materialName, v in pairs(paint[id]) do
			local mat = getMaterial(v.node, 0);
			
			if self.materialNodes[materialName] ~= nil then
				for _, node in pairs(self.materialNodes[materialName]) do
					setMaterial(node, mat, 0);
				end;
			end;

			if self.materialNodes[v.name] ~= nil then
				for _, node in pairs(self.materialNodes[v.name]) do
					setVisibility(node, true);
				end;
			end;
		end;
	end;
end;


-- reload option
function HardPointsControlUnit:developmentReloadFromXML()
	local xmlFile = loadXMLFile("Tmp", self.configFileName);
	
	for _, v in ipairs(self.specializations) do
		if v.developmentReloadFromXML ~= nil then
			v.developmentReloadFromXML(self.environment, xmlFile);
		end;
	end;
	
	delete(xmlFile);
end;


-- checks 
function HardPointsControlUnit.allow(vehicle, adding, object, objectClass) -- check if part is allowed to attach or detach
	local allow, spec, txt = true, object.specs, nil;
	
	for _, v in ipairs(spec) do
		if adding then
			if v.allowAdding ~= nil then
				allow, txt = v.allowAdding(nil, vehicle);
			end;
		else
			if v.allowRemoval ~= nil and objectClass ~= nil then
				allow, txt = v.allowRemoval(objectClass, vehicle);
			end;
		end;
		
		if not allow then
			break;
		end;
	end;
	
	return allow, txt;
end;

function HardPointsControlUnit.canKeepLoading(failedToLoad, tempRootNode, xmlFile) -- "clean" feature for load function
	if failedToLoad then
		if tempRootNode ~= nil then
			delete(tempRootNode);
		end;
		
		if xmlFile ~= nil then
			delete(xmlFile);
		end;
		
		return false;
	else
		return true;
	end;
end;